"""Utilities for processing batches."""


def get_line(batch):
    """Build a list of relevant data to be printed.

    Args:
        batch (dict): an object from project batches response

    Returns:
        str: tab delimited list of relevant data to be printed
    """
    return "\t".join(
        [
            batch["id"],
            batch["last_status"]["created"],
            batch["last_status"]["status"],
            batch["batch_type"],
            batch["name"],
        ]  # noqa
    )



"""Test project list command."""
# pylint: disable=wrong-import-order
import io
import sys
from datetime import datetime, timedelta
from uuid import uuid4

from click import echo
from click.testing import CliRunner

from gencove.client import APIClient, APIClientError  # noqa: I100
from gencove.command.projects.cli import list_projects
from gencove.command.projects.list.constants import (
    PipelineCapabilities,
    Project,
)


def test_list_empty(mocker):
    """Test user organization has no projects."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_projects = mocker.patch.object(
        APIClient,
        "list_projects",
        return_value=dict(results=[], meta=dict(next=None)),
    )
    res = runner.invoke(
        list_projects, ["--email", "foo@bar.com", "--password", "123"]
    )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_get_projects.assert_called_once()
    assert "" in res.output


MOCKED_PROJECTS = dict(
    meta=dict(next=None),
    results=[
        {
            "id": str(uuid4()),
            "name": "test\tproject",
            "description": "",
            "created": (datetime.utcnow() - timedelta(days=7)).isoformat(),
            "organization": str(uuid4()),
            "webhook_url": None,
            "sample_count": 1,
            "pipeline_capabilities": str(uuid4()),
        }
    ],
)

MOCKED_PIPELINE_CAPABILITY = {
    "id": MOCKED_PROJECTS["results"][0]["pipeline_capabilities"],
    "name": "test capability",
    "private": False,
    "merge_vcfs_enabled": False,
}


def test_list_projects_no_permission(mocker):
    """Test projects no permission available to show them."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_projects = mocker.patch.object(
        APIClient,
        "list_projects",
        side_effect=APIClientError(
            message="API Client Error: Not Found: Not found.", status_code=404
        ),
        return_value={"detail": "Not found"},
    )
    mocked_get_pipeline_capabilities = mocker.patch.object(
        APIClient,
        "get_pipeline_capabilities",
        return_value=MOCKED_PIPELINE_CAPABILITY,
    )
    res = runner.invoke(
        list_projects, ["--email", "foo@bar.com", "--password", "123"]
    )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_get_projects.assert_called_once()
    mocked_get_pipeline_capabilities.assert_not_called()

    output_line = io.BytesIO()
    sys.stdout = output_line
    echo(
        "\n".join(
            [
                "ERROR: You do not have permission required to access "
                "the project list.",
                "ERROR: API Client Error: Not Found: Not found.",
                "Aborted!",
            ]
        )
    )
    assert output_line.getvalue() == res.output.encode()


def test_list_projects(mocker):
    """Test projects being outputed to the shell."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_projects = mocker.patch.object(
        APIClient, "list_projects", return_value=MOCKED_PROJECTS
    )
    mocked_get_pipeline_capabilities = mocker.patch.object(
        APIClient,
        "get_pipeline_capabilities",
        return_value=MOCKED_PIPELINE_CAPABILITY,
    )
    res = runner.invoke(
        list_projects, ["--email", "foo@bar.com", "--password", "123"]
    )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_get_projects.assert_called_once()
    mocked_get_pipeline_capabilities.assert_called_once()

    project = Project(**MOCKED_PROJECTS["results"][0])
    pipeline = PipelineCapabilities(**MOCKED_PIPELINE_CAPABILITY)

    output_line = io.BytesIO()
    sys.stdout = output_line
    echo(
        "\t".join(
            [
                project.created,
                project.id,
                project.name.replace("\t", " "),
                pipeline.name,
            ]
        )
    )
    assert output_line.getvalue() == res.output.encode()



"""Commands to be executed from command line."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials

from .constants import UploadOptions
from .main import Upload


@click.command()
@click.argument("source")
@click.argument("destination", required=False)
@add_options(common_options)
@click.option(
    "--run-project-id",
    default=None,
    help="Immediately assign all uploaded files to this project and run them",
)
@click.option(
    "--output",
    required=False,
    default=None,
    help="A destination where to store the resulting assignments.",
)
@click.option(
    "--no-progress",
    is_flag=True,
    help="If specified, no progress bar is shown.",
)
@click.option(
    "--metadata",
    required=False,
    default=None,
    help=(
        "Assign metadata to all samples created from uploads. "
        "Only compatible with --run-project-id."
    ),
)
def upload(  # pylint: disable=E0012,C0330,R0913
    source,
    destination,
    host,
    email,
    password,
    api_key,
    run_project_id,
    output,
    no_progress,
    metadata,
):  # noqa: D301
    """Upload FASTQ files to Gencove's system.

    SOURCE: folder that contains fastq files to be uploaded (acceptable file
    extensions are .fastq.gz, .fastq.bgz, .fq.gz, .fq.bgz), or .fastq-map.csv
    file with the mapping of R1/R2 files to related sample by client_id

    DESTINATION (optional): gncv://[folder], where the folder is the location
    on Gencove systems

    Examples:

        Upload directory contents:

            gencove upload test_dataset gncv://test

    \f

    Args:
        source (.fastq.gz, .fastq.bgz, .fq.gz, .fq.bgz):
            folder that contains fastq files to be uploaded
            OR .fastq-map.csv file
        destination (str, optional): 'gncv://' notated folder
            on Gencove's system, where the files will be uploaded to.
        run_project_id (UUID, optional): ID of a project to which all files
            in this upload will be assigned to and then immediately analyzed.
        output (str, optional): must be used with run_project_id. "-"
            redirects the JSON to STDOUT, and a name redirects the output to
            a file.
        no_progress (bool, optional, default False): do not show progress
            bar.
        metadata (str, optional): JSON metadata to be applied to all samples
    """
    Upload(
        source,
        destination,
        Credentials(email, password, api_key),
        UploadOptions(host, run_project_id, metadata),
        output,
        no_progress,
    ).run()



"""Tests for utils of Gencove CLI."""
import csv
import os

from click.testing import CliRunner

from gencove.command.download.utils import (
    _get_prefix_parts,
    build_file_path,
    get_download_template_format_params,
)
from gencove.command.upload.utils import (
    _validate_header,
    parse_fastqs_map_file,
    upload_file,
)
from gencove.command.utils import is_valid_uuid
from gencove.constants import DOWNLOAD_TEMPLATE, DownloadTemplateParts
from gencove.exceptions import ValidationError


def test_upload_file(mocker):
    """Sanity check upload function."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        with open("foo.txt", "w") as fastq_file:
            fastq_file.write("AAABBB")
        mocked_s3_client = mocker.Mock()
        assert upload_file(
            mocked_s3_client,
            "foo.txt",
            "foo-bucket",
            object_name="foo-object.txt",
        )
        mocked_s3_client.upload_file.assert_called_once()


def test_download_template_tokens():
    """Ensure download tokens are only the ones defined."""
    assert [
        "client_id",
        "gencove_id",
        "file_type",
        "file_extension",
        "default_filename",
    ] == list(DownloadTemplateParts._asdict().values())


# pylint: disable=too-many-locals
def test_build_file_path():
    """Test token combinations when building a file path."""
    client_id = "12345"
    gencove_id = "1"
    deliverable = {
        "download_url": "https://example.com/file.txt",
        "file_type": "txt",
    }
    # file with prefix will never have client_id and gencove_id. 0 is default
    file_with_prefix0 = "{client_id}/{gencove_id}/{default_filename}".format(
        **get_download_template_format_params(client_id, gencove_id)
    )
    file_with_prefix1 = "{client_id}".format(
        **get_download_template_format_params(client_id, gencove_id)
    )
    file_with_prefix2 = "{gencove_id}".format(
        **get_download_template_format_params(client_id, gencove_id)
    )
    file_with_prefix3 = "{file_type}"
    file_with_prefix4 = "{file_extension}"
    file_with_prefix5 = "{default_filename}"
    file_with_prefix6 = "{default_filename}.{file_extension}"
    file_with_prefix7 = "{default_filename}.{client_id}".format(
        **get_download_template_format_params(client_id, gencove_id)
    )
    file_with_prefix8 = "{client_id}.{default_filename}".format(
        **get_download_template_format_params(client_id, gencove_id)
    )
    file_with_prefix9 = "{file_extension}_{client_id}".format(
        **get_download_template_format_params(client_id, gencove_id)
    )
    download_to = "."
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = build_file_path(deliverable, file_with_prefix0, download_to)
        # default_filename is always a full filename with any extensions
        assert result == "./{}/{}/{}".format(
            client_id, gencove_id, "file.txt"
        )
        result = build_file_path(deliverable, file_with_prefix1, download_to)
        assert result == "./{}".format(client_id)
        result = build_file_path(deliverable, file_with_prefix2, download_to)
        assert result == "./{}".format(gencove_id)
        result = build_file_path(deliverable, file_with_prefix3, download_to)
        assert result == "./{}".format("txt")
        result = build_file_path(deliverable, file_with_prefix4, download_to)
        assert result == "./{}".format("txt")
        result = build_file_path(deliverable, file_with_prefix5, download_to)
        assert result == "./{}".format("file.txt")
        result = build_file_path(deliverable, file_with_prefix6, download_to)
        assert result == "./{}.{}".format("file.txt", "txt")
        result = build_file_path(deliverable, file_with_prefix7, download_to)
        assert result == "./{}.{}".format("file.txt", client_id)
        result = build_file_path(deliverable, file_with_prefix8, download_to)
        assert result == "./{}.{}".format(client_id, "file.txt")
        result = build_file_path(deliverable, file_with_prefix9, download_to)
        assert result == "./{}_{}".format("txt", client_id)


def test___get_filename_dirs_prefix():
    """Test proper processing of parts in download template."""
    client_id = "12345"
    gencove_id = "1"
    template = DOWNLOAD_TEMPLATE.format(
        **get_download_template_format_params(client_id, gencove_id)
    )

    resp = _get_prefix_parts(template)

    assert resp.dirs == "{}/{}".format(client_id, gencove_id)
    assert resp.filename == "{{{}}}".format(
        DownloadTemplateParts.default_filename
    )
    assert resp.file_extension == ""

    template2 = "{client_id}-{gencove_id}_{file_type}".format(
        **get_download_template_format_params(client_id, gencove_id)
    )
    resp = _get_prefix_parts(template2)
    assert resp.dirs == ""
    assert resp.filename == "{}-{}_{{{}}}".format(
        client_id, gencove_id, DownloadTemplateParts.file_type
    )
    assert resp.file_extension == ""

    template3 = "{client_id}-{gencove_id}_{file_type}.vcf.gz".format(
        **get_download_template_format_params(client_id, gencove_id)
    )
    resp = _get_prefix_parts(template3)
    assert resp.dirs == ""
    assert resp.filename == "{}-{}_{{{}}}".format(
        client_id, gencove_id, DownloadTemplateParts.file_type
    )
    assert resp.file_extension == "vcf.gz"


def test_parse_fastqs_map_file():
    """Test parsing of map file into dict."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("test_dir")
        with open("test_dir/test.fastq.gz", "w") as fastq_file1:
            fastq_file1.write("AAABBB")

        with open("test_dir/test_2.fastq.gz", "w") as fastq_file2:
            fastq_file2.write("AAABBB")

        with open("test_dir/test_3.fastq.gz", "w") as fastq_file3:
            fastq_file3.write("AAABBB")

        with open("test_map.csv", "w") as map_file:
            writer = csv.writer(map_file)
            writer.writerows(
                [
                    ["client_id", "r_notation", "path"],
                    ["barid", "r1", "test_dir/test.fastq.gz"],
                    ["barid", "r2", "test_dir/test_2.fastq.gz"],
                    ["barid", "r1", "test_dir/test_3.fastq.gz"],
                ]
            )

        fastqs = parse_fastqs_map_file("test_map.csv")
        assert len(fastqs) == 2
        assert ("barid", "R1") in fastqs
        assert ("barid", "R2") in fastqs
        assert len(fastqs[("barid", "R1")]) == 2
        assert fastqs[("barid", "R1")][0] == "test_dir/test.fastq.gz"
        assert fastqs[("barid", "R1")][1] == "test_dir/test_3.fastq.gz"


def test_invalid_fastqs_map_file():
    """Test that error is raised if file is invalid."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("test_dir")
        with open("test_dir/test.fastq.gz", "w") as fastq_file1:
            fastq_file1.write("AAABBB")

        with open("test_dir/test_2.fastq.gz", "w") as fastq_file2:
            fastq_file2.write("AAABBB")

        with open("test_dir/test_3.fastq.gz", "w") as fastq_file3:
            fastq_file3.write("AAABBB")

        with open("test_map.csv", "w") as map_file:
            writer = csv.writer(map_file)
            writer.writerows(
                [
                    ["client_id", "r_notation", "path"],
                    ["barid", "r1", "test_dir/test.fastq.gz"],
                    ["barid", "r2", "test_dir/test_2.fastq.zip"],
                    ["barid", "r1", "test_dir/test_3.fastq.gz"],
                ]
            )
        try:
            parse_fastqs_map_file("test_map.csv")
        except ValidationError as err:
            assert "Bad file extension in path" in err.args[0]


def test_fastqs_map_file_path_does_not_exist():
    """Test that error is raised if file is invalid."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("test_dir")
        with open("test_dir/test.fastq.gz", "w") as fastq_file1:
            fastq_file1.write("AAABBB")

        with open("test_dir/test_2.fastq.gz", "w") as fastq_file2:
            fastq_file2.write("AAABBB")

        with open("test_dir/test_3.fastq.gz", "w") as fastq_file3:
            fastq_file3.write("AAABBB")

        with open("test_map.csv", "w") as map_file:
            writer = csv.writer(map_file)
            writer.writerows(
                [
                    ["client_id", "r_notation", "path"],
                    ["barid", "r1", "test_dir/test.fastq.gz"],
                    ["barid", "r1", "test_dir/test_3.fastq.gz"],
                    ["barid", "r1", "test_dir/i_dont_exist.fastq.gz"],
                ]
            )
        try:
            parse_fastqs_map_file("test_map.csv")
        except ValidationError as err:
            assert "Could not find" in err.args[0]


def test_fastqs_map_file_has_wrong_header():
    """Test that header is validated properly."""
    with open("test_map.csv", "w") as map_file:
        writer = csv.writer(map_file)
        writer.writerows(
            [
                ["client_id", "r1_notation", "path"],
                ["barid", "r1", "test_dir/test.fastq.gz"],
                ["barid", "r1", "test_dir/test_3.fastq.gz"],
                ["barid", "r1", "test_dir/i_dont_exist.fastq.gz"],
            ]
        )
    try:
        parse_fastqs_map_file("test_map.csv")
    except ValidationError as err:
        assert "Unexpected CSV header" in err.args[0]


def test__validate_header():
    """Test that header is validated properly."""
    header_row = dict(foo="foo", bar="bar")
    try:
        _validate_header(header_row)
    except ValidationError:
        pass

    header_row = dict(
        client_id="client_id", r_notation="r_notation", path="path"
    )
    assert _validate_header(header_row) is None


def test_is_valid_uuid__is_valid():
    """"Test that a UUID is a valid UUID"""
    assert is_valid_uuid("11111111-1111-1111-1111-111111111111")


def test_is_valid_uuid__is_not_valid__too_long():
    """"Test that UUID with extra chars is not a valid UUID"""
    assert is_valid_uuid("11111111-1111-1111-1111-11111111111122") is False


def test_is_valid_uuid__is_not_valid__too_short():
    """"Test that UUID with missing chars is not a valid UUID"""
    assert is_valid_uuid("11111111-1111-1111-1111-1") is False


def test_is_valid_uuid__is_not_valid__text():
    """"Test that random word is not a valid UUID"""
    assert is_valid_uuid("foo") is False



"""Samples list shell command definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import (
    Credentials,
    SAMPLE_ARCHIVE_STATUS,
    SAMPLE_STATUS,
)

from .constants import SamplesOptions
from .main import ListSamples


@click.command("list-samples")
@click.argument("project_id")
@click.option(
    "--search", help="Gencove sample ID, client ID or metadata substring"
)
@click.option(
    "--status",
    help="Get samples with specific status",
    type=click.Choice(SAMPLE_STATUS._asdict().values()),
    default=SAMPLE_STATUS.all,
)
@click.option(
    "--archive-status",
    help="Get samples with specific archive status",
    type=click.Choice(SAMPLE_ARCHIVE_STATUS._asdict().values()),
    default=SAMPLE_ARCHIVE_STATUS.all,
)
@add_options(common_options)
def list_project_samples(  # pylint: disable=E0012,C0330,R0913
    project_id, search, status, archive_status, host, email, password, api_key
):
    """List samples in a project."""
    ListSamples(
        project_id,
        Credentials(email, password, api_key),
        SamplesOptions(host, status, archive_status, search),
    ).run()



"""List projects shell command definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials, Optionals

from .main import List


@click.command("list")
@add_options(common_options)
def list_projects(host, email, password, api_key):
    """List your projects."""
    List(Credentials(email, password, api_key), Optionals(host)).run()



"""Test download command."""
import io
import json
import os
import sys
from uuid import uuid4

from click import echo
from click.testing import CliRunner

from gencove.cli import download
from gencove.client import APIClient


def test_no_required_options():
    """Test that command exits without project id or sample id provided."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        res = runner.invoke(
            download,
            [
                "cli_test_data",
                "--project-id",
                "123",
                "--email",
                "foo@bar.com",
                "--password",
                "12345",
            ],
        )
        assert res.exit_code == 1


def test_both_project_id_and_sample_ids_provided():
    """Command exits if both project id and sample ids are provided."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("cli_test_data")
        res = runner.invoke(
            download,
            [
                "cli_test_data",
                "--project-id",
                "123",
                "--sample-ids",
                "1,2,3",
                "--email",
                "foo@bar.com",
                "--password",
                "12345",
            ],
        )
    assert res.exit_code == 1


def test_project_id_provided(mocker):
    """Check happy flow."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        mocked_login = mocker.patch.object(
            APIClient, "login", return_value=None
        )
        mocked_project_samples = mocker.patch.object(
            APIClient,
            "get_project_samples",
            return_value={"results": [{"id": 0}], "meta": {"next": None}},
        )
        mocked_sample_details = mocker.patch.object(
            APIClient,
            "get_sample_details",
            return_value={
                "id": 0,
                "client_id": 1,
                "last_status": {
                    "id": str(uuid4()),
                    "status": "succeeded",
                    "created": "2020-07-28T12:46:22.719862Z",
                },
                "archive_last_status": {
                    "id": str(uuid4()),
                    "status": "available",
                    "created": "2020-07-28T12:46:22.719862Z",
                    "transition_cutoff": "2020-08-28T12:46:22.719862Z",
                },
                "files": [
                    {
                        "id": str(uuid4()),
                        "file_type": "txt",
                        "download_url": "https://foo.com/bar.txt",
                    }
                ],
            },
        )
        mocked_qc_metrics = mocker.patch.object(
            APIClient,
            "get_sample_qc_metrics",
            return_value={"results": [{"foo": 12}]},
        )
        mocked_get_metadata = mocker.patch.object(
            APIClient,
            "get_metadata",
            return_value=dict(
                metadata=None,
            ),
        )
        mocked_download_file = mocker.patch(
            "gencove.command.download.main.download_file"
        )
        res = runner.invoke(
            download,
            [
                "cli_test_data",
                "--project-id",
                "123",
                "--email",
                "foo@bar.com",
                "--password",
                "123",
            ],
        )
        assert res.exit_code == 0
        mocked_login.assert_called_once()
        mocked_project_samples.assert_called_once()
        mocked_qc_metrics.assert_called_once()
        mocked_get_metadata.assert_called_once()
        mocked_download_file.assert_called_once()
        mocked_sample_details.assert_called_once()


def test_sample_ids_provided(mocker):
    """Check happy flow with sample ids."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        mocked_login = mocker.patch.object(
            APIClient, "login", return_value=None
        )
        mocked_sample_details = mocker.patch.object(
            APIClient,
            "get_sample_details",
            return_value={
                "id": 0,
                "client_id": 1,
                "last_status": {
                    "id": str(uuid4()),
                    "status": "succeeded",
                    "created": "2020-07-28T12:46:22.719862Z",
                },
                "archive_last_status": {
                    "id": str(uuid4()),
                    "status": "available",
                    "created": "2020-07-28T12:46:22.719862Z",
                    "transition_cutoff": "2020-08-28T12:46:22.719862Z",
                },
                "files": [
                    {
                        "id": str(uuid4()),
                        "file_type": "txt",
                        "download_url": "https://foo.com/bar.txt",
                    }
                ],
            },
        )
        mocked_qc_metrics = mocker.patch.object(
            APIClient,
            "get_sample_qc_metrics",
            return_value={"results": [{"foo": 12}]},
        )
        mocked_get_metadata = mocker.patch.object(
            APIClient,
            "get_metadata",
            return_value=dict(
                metadata=None,
            ),
        )
        mocked_download_file = mocker.patch(
            "gencove.command.download.main.download_file"
        )
        res = runner.invoke(
            download,
            [
                "cli_test_data",
                "--sample-ids",
                "0",
                "--email",
                "foo@bar.com",
                "--password",
                "123",
            ],
        )
        assert res.exit_code == 0
        mocked_login.assert_called_once()
        mocked_download_file.assert_called_once_with(
            "cli_test_data/1/0/bar.txt",
            "https://foo.com/bar.txt",
            True,
            False,
        )
        mocked_sample_details.assert_called_once()
        mocked_qc_metrics.assert_called_once()
        mocked_get_metadata.assert_called_once()


def test_sample_ids_provided_no_qc_file(mocker):
    """Check flow with sample ids and no file present."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        mocked_login = mocker.patch.object(
            APIClient, "login", return_value=None
        )
        mocked_sample_details = mocker.patch.object(
            APIClient,
            "get_sample_details",
            return_value={
                "id": 0,
                "client_id": 1,
                "last_status": {
                    "id": str(uuid4()),
                    "status": "succeeded",
                    "created": "2020-07-28T12:46:22.719862Z",
                },
                "archive_last_status": {
                    "id": str(uuid4()),
                    "status": "available",
                    "created": "2020-07-28T12:46:22.719862Z",
                    "transition_cutoff": "2020-08-28T12:46:22.719862Z",
                },
                "files": [],
            },
        )
        mocked_qc_metrics = mocker.patch.object(
            APIClient,
            "get_sample_qc_metrics",
            return_value={"results": [{"foo": 12}]},
        )
        mocked_get_metadata = mocker.patch.object(
            APIClient,
            "get_metadata",
            return_value={"metadata": [{"foo": 12}]},
        )
        res = runner.invoke(
            download,
            [
                "cli_test_data",
                "--sample-ids",
                "0",
                "--email",
                "foo@bar.com",
                "--password",
                "123",
            ],
        )
        assert res.exit_code == 0
        mocked_login.assert_called_once()
        mocked_sample_details.assert_called_once()
        mocked_qc_metrics.assert_called_once()
        mocked_get_metadata.assert_called_once()


def test_sample_ids_provided_no_metadata_file(mocker):
    """Check flow with sample ids and no metadata present."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        mocked_login = mocker.patch.object(
            APIClient, "login", return_value=None
        )
        mocked_sample_details = mocker.patch.object(
            APIClient,
            "get_sample_details",
            return_value={
                "id": 0,
                "client_id": 1,
                "last_status": {
                    "id": str(uuid4()),
                    "status": "succeeded",
                    "created": "2020-07-28T12:46:22.719862Z",
                },
                "archive_last_status": {
                    "id": str(uuid4()),
                    "status": "available",
                    "created": "2020-07-28T12:46:22.719862Z",
                    "transition_cutoff": "2020-08-28T12:46:22.719862Z",
                },
                "files": [],
            },
        )
        mocked_get_metadata = mocker.patch.object(
            APIClient,
            "get_metadata",
            return_value={"metadata": [{"foo": 12}]},
        )
        mocked_qc_metrics = mocker.patch.object(
            APIClient,
            "get_sample_qc_metrics",
            return_value={"results": [{"foo": 12}]},
        )
        res = runner.invoke(
            download,
            [
                "cli_test_data",
                "--sample-ids",
                "0",
                "--email",
                "foo@bar.com",
                "--password",
                "123",
            ],
        )
        assert res.exit_code == 0
        mocked_login.assert_called_once()
        mocked_sample_details.assert_called_once()
        mocked_get_metadata.assert_called_once()
        mocked_qc_metrics.assert_called_once()


def test_multiple_credentials_not_allowed():
    """Test that in providing multiple credentials is not allowed."""
    runner = CliRunner()
    res = runner.invoke(
        download,
        [
            "some_test_data",
            "--sample-ids",
            "0",
            "--email",
            "foo@bar.com",
            "--password",
            "123",
            "--api-key",
            "foobar",
        ],
    )
    assert res.exit_code == 1
    assert "Please provide either username/password or API key." in res.output


def test_download_stdout_no_flag(mocker):
    """Test command exits if no flag provided and stdout defined."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_project_samples = mocker.patch.object(
        APIClient,
        "get_project_samples",
        return_value={"results": [{"id": 0}], "meta": {"next": None}},
    )
    res = runner.invoke(
        download,
        [
            "-",
            "--project-id",
            "123",
            "--email",
            "foo@bar.com",
            "--password",
            "12345",
        ],
    )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_project_samples.assert_called_once()
    output_line = io.BytesIO()
    sys.stdout = output_line
    echo("ERROR: Cannot have - as a destination without download-urls.")
    assert output_line.getvalue() in res.output.encode()


def test_download_stdout_with_flag(mocker):
    """Test command outputs json to stdout."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_project_samples = mocker.patch.object(
        APIClient,
        "get_project_samples",
        return_value={"results": [{"id": 0}], "meta": {"next": None}},
    )
    last_status_id = str(uuid4())
    archive_last_status_id = str(uuid4())
    file_id = str(uuid4())
    mocked_sample_details = mocker.patch.object(
        APIClient,
        "get_sample_details",
        return_value={
            "id": 0,
            "client_id": 1,
            "last_status": {
                "id": last_status_id,
                "status": "succeeded",
                "created": "2020-07-28T12:46:22.719862Z",
            },
            "archive_last_status": {
                "id": archive_last_status_id,
                "status": "available",
                "created": "2020-07-28T12:46:22.719862Z",
                "transition_cutoff": "2020-08-28T12:46:22.719862Z",
            },
            "files": [
                {
                    "id": file_id,
                    "file_type": "txt",
                    "download_url": "https://foo.com/bar.txt",
                }
            ],
        },
    )
    res = runner.invoke(
        download,
        [
            "-",
            "--project-id",
            "123",
            "--email",
            "foo@bar.com",
            "--password",
            "12345",
            "--download-urls",
        ],
    )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_project_samples.assert_called_once()
    mocked_sample_details.assert_called_once()
    output_line = io.BytesIO()
    sys.stdout = output_line
    mocked_result = json.dumps(
        [
            {
                "gencove_id": 0,
                "client_id": 1,
                "last_status": {
                    "id": last_status_id,
                    "status": "succeeded",
                    "created": "2020-07-28T12:46:22.719862Z",
                },
                "archive_last_status": {
                    "id": archive_last_status_id,
                    "status": "available",
                    "created": "2020-07-28T12:46:22.719862Z",
                    "transition_cutoff": "2020-08-28T12:46:22.719862Z",
                },
                "files": {
                    "txt": {
                        "id": file_id,
                        "download_url": "https://foo.com/bar.txt",
                    }
                },
            }
        ],
        indent=4,
    )
    echo(mocked_result)
    assert output_line.getvalue() in res.output.encode()


def test_download_urls_to_file(mocker):
    """Test saving downloaded urls output to a json file."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_project_samples = mocker.patch.object(
        APIClient,
        "get_project_samples",
        return_value={"results": [{"id": 0}], "meta": {"next": None}},
    )
    last_status_id = str(uuid4())
    archive_last_status_id = str(uuid4())
    file_id = str(uuid4())
    mocked_sample_details = mocker.patch.object(
        APIClient,
        "get_sample_details",
        return_value={
            "id": 0,
            "client_id": 1,
            "last_status": {
                "id": last_status_id,
                "status": "succeeded",
                "created": "2020-07-28T12:46:22.719862Z",
            },
            "archive_last_status": {
                "id": archive_last_status_id,
                "status": "available",
                "created": "2020-07-28T12:46:22.719862Z",
                "transition_cutoff": "2020-08-28T12:46:22.719862Z",
            },
            "files": [
                {
                    "id": file_id,
                    "file_type": "txt",
                    "download_url": "https://foo.com/bar.txt",
                }
            ],
        },
    )
    mocked_output_list = mocker.patch(
        "gencove.command.download.main.Download.output_list"
    )
    with runner.isolated_filesystem():
        res = runner.invoke(
            download,
            [
                "output.json",
                "--project-id",
                "123",
                "--email",
                "foo@bar.com",
                "--password",
                "12345",
                "--download-urls",
            ],
        )
        assert res.exit_code == 0
        mocked_login.assert_called_once()
        mocked_project_samples.assert_called_once()
        mocked_sample_details.assert_called_once()
        mocked_output_list.assert_called_once()


def test_download_no_progress(mocker):
    """Test command doesn't show progress bar."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        mocked_login = mocker.patch.object(
            APIClient, "login", return_value=None
        )
        mocked_sample_details = mocker.patch.object(
            APIClient,
            "get_sample_details",
            return_value={
                "id": 0,
                "client_id": 1,
                "last_status": {
                    "id": str(uuid4()),
                    "status": "succeeded",
                    "created": "2020-07-28T12:46:22.719862Z",
                },
                "archive_last_status": {
                    "id": str(uuid4()),
                    "status": "available",
                    "created": "2020-07-28T12:46:22.719862Z",
                    "transition_cutoff": "2020-08-28T12:46:22.719862Z",
                },
                "files": [
                    {
                        "id": str(uuid4()),
                        "file_type": "txt",
                        "download_url": "https://foo.com/bar.txt",
                    }
                ],
            },
        )
        mocked_qc_metrics = mocker.patch.object(
            APIClient,
            "get_sample_qc_metrics",
            return_value={"results": [{"foo": 12}]},
        )
        mocked_get_metadata = mocker.patch.object(
            APIClient,
            "get_metadata",
            return_value=dict(
                metadata=None,
            ),
        )
        mocked_download_file = mocker.patch(
            "gencove.command.download.main.download_file"
        )
        res = runner.invoke(
            download,
            [
                "cli_test_data",
                "--sample-ids",
                "0",
                "--email",
                "foo@bar.com",
                "--password",
                "123",
                "--no-progress",
            ],
        )
        assert res.exit_code == 0
        mocked_login.assert_called_once()
        mocked_download_file.assert_called_once_with(
            "cli_test_data/1/0/bar.txt", "https://foo.com/bar.txt", True, True
        )
        mocked_sample_details.assert_called_once()
        mocked_qc_metrics.assert_called_once()
        mocked_get_metadata.assert_called_once()


def test_project_id_provided_skip_existing_qc_and_metadata(mocker):
    """Check happy flow."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        mocked_login = mocker.patch.object(
            APIClient, "login", return_value=None
        )
        mocked_project_samples = mocker.patch.object(
            APIClient,
            "get_project_samples",
            return_value={"results": [{"id": 0}], "meta": {"next": None}},
        )
        mocked_sample_details = mocker.patch.object(
            APIClient,
            "get_sample_details",
            return_value={
                "id": 0,
                "client_id": 1,
                "last_status": {
                    "id": str(uuid4()),
                    "status": "succeeded",
                    "created": "2020-07-28T12:46:22.719862Z",
                },
                "archive_last_status": {
                    "id": str(uuid4()),
                    "status": "available",
                    "created": "2020-07-28T12:46:22.719862Z",
                    "transition_cutoff": "2020-08-28T12:46:22.719862Z",
                },
                "files": [
                    {
                        "id": str(uuid4()),
                        "file_type": "txt",
                        "download_url": "https://foo.com/bar.txt",
                    }
                ],
            },
        )
        mocked_qc_metrics = mocker.patch.object(
            APIClient,
            "get_sample_qc_metrics",
            return_value={"results": [{"foo": 12}]},
        )
        mocked_get_metadata = mocker.patch.object(
            APIClient,
            "get_metadata",
            return_value=dict(
                metadata=None,
            ),
        )
        mocked_download_file = mocker.patch(
            "gencove.command.download.main.download_file"
        )
        res = runner.invoke(
            download,
            [
                "cli_test_data",
                "--project-id",
                "123",
                "--email",
                "foo@bar.com",
                "--password",
                "123",
                "--no-skip-existing",
            ],
        )
        assert res.exit_code == 0
        mocked_login.assert_called_once()
        mocked_project_samples.assert_called_once()
        mocked_qc_metrics.assert_called_once()
        mocked_get_metadata.assert_called_once()
        mocked_download_file.assert_called_once()
        mocked_sample_details.assert_called_once()

        # call it the second time
        mocked_qc_metrics = mocker.patch.object(
            APIClient,
            "get_sample_qc_metrics",
            return_value={"results": [{"foo": 12}]},
        )
        mocked_get_metadata = mocker.patch.object(
            APIClient,
            "get_metadata",
            return_value=dict(
                metadata=None,
            ),
        )
        res = runner.invoke(
            download,
            [
                "cli_test_data",
                "--project-id",
                "123",
                "--email",
                "foo@bar.com",
                "--password",
                "123",
                "--skip-existing",
            ],
        )
        assert res.exit_code == 0
        mocked_qc_metrics.assert_not_called()
        mocked_get_metadata.assert_not_called()


def test_download_not_working_because_archived(mocker):
    """Test command doesn't download archived files."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        mocked_login = mocker.patch.object(
            APIClient, "login", return_value=None
        )
        mocked_sample_details = mocker.patch.object(
            APIClient,
            "get_sample_details",
            return_value={
                "id": 0,
                "client_id": 1,
                "last_status": {
                    "id": str(uuid4()),
                    "status": "succeeded",
                    "created": "2020-07-28T12:46:22.719862Z",
                },
                "archive_last_status": {
                    "id": str(uuid4()),
                    "status": "archived",
                    "created": "2020-07-28T12:46:22.719862Z",
                    "transition_cutoff": "2020-08-28T12:46:22.719862Z",
                },
                "files": [
                    {
                        "id": str(uuid4()),
                        "file_type": "txt",
                        "download_url": None,
                    }
                ],
            },
        )
        mocked_qc_metrics = mocker.patch.object(
            APIClient,
            "get_sample_qc_metrics",
            return_value={"results": [{"foo": 12}]},
        )
        mocked_get_metadata = mocker.patch.object(
            APIClient,
            "get_metadata",
            return_value=dict(
                metadata=None,
            ),
        )
        mocked_download_file = mocker.patch(
            "gencove.command.download.main.download_file"
        )
        res = runner.invoke(
            download,
            [
                "cli_test_data",
                "--sample-ids",
                "0",
                "--email",
                "foo@bar.com",
                "--password",
                "123",
            ],
        )
        assert res.exit_code == 1
        mocked_login.assert_called_once()
        mocked_download_file.assert_not_called()
        mocked_sample_details.assert_called_once()
        mocked_qc_metrics.assert_not_called()
        mocked_get_metadata.assert_not_called()


def test_project_id_provided_filter_not_archived(mocker):
    """Check happy flow."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        mocked_login = mocker.patch.object(
            APIClient, "login", return_value=None
        )
        mocked_project_samples = mocker.patch.object(
            APIClient,
            "get_project_samples",
            return_value={"results": [], "meta": {"next": None}},
        )
        mocked_sample_details = mocker.patch.object(
            APIClient,
            "get_sample_details",
            return_value={
                "id": 0,
                "client_id": 1,
                "last_status": {
                    "id": str(uuid4()),
                    "status": "succeeded",
                    "created": "2020-07-28T12:46:22.719862Z",
                },
                "archive_last_status": {
                    "id": str(uuid4()),
                    "status": "archived",
                    "created": "2020-07-28T12:46:22.719862Z",
                    "transition_cutoff": "2020-08-28T12:46:22.719862Z",
                },
                "files": [
                    {
                        "id": str(uuid4()),
                        "file_type": "txt",
                        "download_url": None,
                    }
                ],
            },
        )
        mocked_qc_metrics = mocker.patch.object(
            APIClient,
            "get_sample_qc_metrics",
            return_value={"results": [{"foo": 12}]},
        )
        mocked_get_metadata = mocker.patch.object(
            APIClient,
            "get_metadata",
            return_value=dict(
                metadata=None,
            ),
        )
        mocked_download_file = mocker.patch(
            "gencove.command.download.main.download_file"
        )
        res = runner.invoke(
            download,
            [
                "cli_test_data",
                "--project-id",
                "123",
                "--email",
                "foo@bar.com",
                "--password",
                "123",
            ],
        )
        assert res.exit_code == 1
        mocked_login.assert_called_once()
        mocked_project_samples.assert_called_once()
        mocked_qc_metrics.assert_not_called()
        mocked_get_metadata.assert_not_called()
        mocked_download_file.assert_not_called()
        mocked_sample_details.assert_not_called()



"""Utilities for different levels of click.echo outputs."""
import os
from datetime import datetime

import click


INFO = "INFO"
DEBUG = "DEBUG"
LOG_LEVEL = os.environ.get("GENCOVE_LOG", INFO)


def _echo_with_datetime(msg, **kwargs):
    """Output click echo message with datetime."""
    click.echo("{}  {}".format(datetime.utcnow().isoformat(), msg), **kwargs)


def _echo(msg, **kwargs):
    """Output click echo message without datetime."""
    click.echo(msg, **kwargs)


if LOG_LEVEL == DEBUG:
    import sys
    import logging
    from platform import platform
    import boto3
    from gencove.version import version

    _echo_with_datetime(
        "Python version: {}.{}.{}".format(
            sys.version_info.major,
            sys.version_info.minor,
            sys.version_info.micro,
        )
    )
    _echo_with_datetime("CLI version: {}".format(version()))
    _echo_with_datetime("OS details: {}".format(platform()))
    _echo_with_datetime("boto3 version: {}".format(boto3.__version__))
    logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
    logging.getLogger("botocore.auth").setLevel(logging.CRITICAL)


def output_warning(text):
    """Click echo warning."""
    return click.style(
        "WARNING: {}".format(text), fg="bright_yellow", bold=True
    )


def output_error(text):
    """Click echo error."""
    return click.style("ERROR: {}".format(text), fg="bright_red", bold=True)


def echo_data(msg, **kwargs):
    """Output click echo msg."""
    if LOG_LEVEL == DEBUG:
        _echo_with_datetime(msg, **kwargs)
    else:
        _echo(msg, **kwargs)


def echo_info(msg, **kwargs):
    """Output click echo msg."""
    if LOG_LEVEL == DEBUG:
        _echo_with_datetime(msg, err=True, **kwargs)
    else:
        _echo(msg, err=True, **kwargs)


def echo_debug(msg, **kwargs):
    """Output click echo msg only if debug is on."""
    if LOG_LEVEL == DEBUG:
        _echo_with_datetime(msg, err=True, **kwargs)


def echo_warning(msg, **kwargs):
    """Output click echo msg with background."""
    if LOG_LEVEL == DEBUG:
        _echo_with_datetime(output_warning(msg), err=True, **kwargs)
    else:
        _echo(output_warning(msg), err=True, **kwargs)


def echo_error(msg, **kwargs):
    """Output click echo msg with background."""
    if LOG_LEVEL == DEBUG:
        _echo_with_datetime(output_error(msg), err=True, **kwargs)
    else:
        _echo(output_error(msg), err=True, **kwargs)



"""Gencove CLI utils."""
import os
import re

import boto3

from botocore.credentials import RefreshableCredentials
from botocore.session import get_session

import click

import progressbar

from gencove.client import APIClientError  # noqa: I100
from gencove.logger import echo_debug, echo_error, echo_info, echo_warning

KB = 1024
MB = KB * 1024
GB = MB * 1024
NUM_MB_IN_CHUNK = 100
CHUNK_SIZE = NUM_MB_IN_CHUNK * MB
FILENAME_RE = re.compile("filename=(.+)")


def get_s3_client_refreshable(refresh_method):
    """Return thread-safe s3 client with refreshable credentials.

    :param refresh_method: function that can get fresh credentials
    """
    session = get_session()
    session_credentials = RefreshableCredentials.create_from_metadata(
        metadata=refresh_method(),
        refresh_using=refresh_method,
        method="sts-assume-role",
    )
    # pylint: disable=protected-access
    session._credentials = session_credentials
    boto3_session = boto3.Session(botocore_session=session)
    return boto3_session.client(
        "s3",
        endpoint_url=os.environ.get("GENCOVE_LOCALSTACK_S3_ENDPOINT") or None,
    )


def get_progress_bar(total_size, action):
    """Get progressbar.ProgressBar instance for file transfer.

    Args:
        total_size: int
        action: str that will be prepended to the progressbar.
            i.e "Uploading: " or "Downloading: "

    Returns:
        progressbar.ProgressBar instance
    """
    return progressbar.ProgressBar(
        max_value=total_size,
        widgets=[
            action,
            progressbar.Percentage(),
            " ",
            progressbar.Bar(marker="#", left="[", right="]"),
            " ",
            progressbar.ETA(),
            " ",
            progressbar.Timer(),
            " ",
            progressbar.FileTransferSpeed(),
        ],
        redirect_stdout=True,
    )


def get_regular_progress_bar(total_size, action):
    """Get progressbar.ProgressBar instance.

    Args:
        total_size: int
        action: str that will be prepended to the progressbar.
            i.e "Uploading: " or "Downloading: "

    Returns:
        progressbar.ProgressBar instance
    """
    return progressbar.ProgressBar(
        max_value=total_size,
        redirect_stdout=True,
        widgets=[
            action,
            progressbar.Percentage(),
            " ",
            progressbar.Bar(marker="#", left="[", right="]"),
            " ",
            progressbar.ETA(),
            " ",
            progressbar.Timer(),
        ],
    )


def validate_credentials(credentials):
    """Validate user credentials."""
    if credentials.email and credentials.password and credentials.api_key:
        echo_debug("User provided 2 sets of credentials.")
        echo_warning(
            "Multiple sets of credentials provided."
            "Please provide either username/password or API key."
        )
        return False

    return True


def login(api_client, credentials):
    """Login user into Gencove's system."""
    if credentials.api_key:
        echo_debug("User authorized via api key")
        api_client.set_api_key(credentials.api_key)
        return True

    email = credentials.email
    password = credentials.password

    if not credentials.email or not credentials.password:
        echo_info("Login required")
        email = email or click.prompt("Email", type=str, err=True)
        password = password or click.prompt(
            "Password", type=str, hide_input=True, err=True
        )

    try:
        api_client.login(email, password)
        echo_debug("User logged in successfully")
        return True
    except APIClientError as err:
        echo_debug("Failed to login: {}".format(err))
        echo_error(
            "Failed to login. Please verify your credentials and try again"
        )
        return False


def batchify(items_list, batch_size=500):
    """Generate batches from items list.

    Args:
        items_list (list): list that will be batchified.
        batch_size (int, default=500): batch size that will be returned.
            last batch is not promised to be exactly the length of batch_size.

    Returns:
        subset of items_list
    """
    total = len(items_list)
    left_to_process = total
    start = 0
    while left_to_process >= 0:
        end = start + batch_size
        if end > total:
            end = total
        yield items_list[start:end]
        start += batch_size
        left_to_process -= batch_size



"""Test uploads list command."""
# pylint: disable=wrong-import-order
import io
import sys
from uuid import uuid4

from click import echo
from click.testing import CliRunner

from gencove.client import APIClient, APIClientError
from gencove.command.uploads.list.cli import list_uploads


def test_list_does_not_exist(mocker):
    """Test user cannot get to uploads."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_projects = mocker.patch.object(
        APIClient,
        "get_sample_sheet",
        side_effect=APIClientError(
            message="API Client Error: Not Found: Not found.", status_code=404
        ),
        return_value={"detail": "Not found"},
    )
    res = runner.invoke(
        list_uploads, ["--email", "foo@bar.com", "--password", "123"]
    )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_get_projects.assert_called_once()
    assert (
        "\n".join(
            [
                "ERROR: Uploads do not exist or you do not have permission "
                "required to access them.",
                "ERROR: API Client Error: Not Found: Not found.",
                "Aborted!\n",
            ]
        )
        == res.output
    )


def test_list_empty(mocker):
    """Test user has no uploads."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_projects = mocker.patch.object(
        APIClient,
        "get_sample_sheet",
        return_value=dict(results=[], meta=dict(next=None)),
    )
    res = runner.invoke(
        list_uploads, ["--email", "foo@bar.com", "--password", "123"]
    )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_get_projects.assert_called_once()
    assert "" in res.output


MOCKED_UPLOADS = dict(
    meta=dict(next=None),
    results=[
        {
            "client_id": "clientid1",
            "fastq": {
                "r1": {
                    "upload": str(uuid4()),
                    "destination_path": "gncv://batch1/clientid1_R1.fastq.gz",
                },
                "r2": {
                    "upload": str(uuid4()),
                    "destination_path": "gncv://batch1/clientid1_R@.fastq.gz",
                },
            },
        },
        {
            "client_id": "clientid2",
            "fastq": {
                "r1": {
                    "upload": str(uuid4()),
                    "destination_path": "gncv://batch2/clientid2_R1.fastq.gz",
                }
            },
        },
    ],
)


def test_list_uploads(mocker):
    """Test list uploads being outputed to the shell."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_sample_sheet = mocker.patch.object(
        APIClient, "get_sample_sheet", return_value=MOCKED_UPLOADS
    )
    res = runner.invoke(
        list_uploads, ["--email", "foo@bar.com", "--password", "123"]
    )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_get_sample_sheet.assert_called_once()

    output_line = io.BytesIO()
    sys.stdout = output_line
    echo(
        "\t".join(
            [
                MOCKED_UPLOADS["results"][0]["client_id"],
                MOCKED_UPLOADS["results"][0]["fastq"]["r1"]["upload"],
                MOCKED_UPLOADS["results"][0]["fastq"]["r1"][
                    "destination_path"
                ],
                MOCKED_UPLOADS["results"][0]["fastq"]["r2"]["upload"],
                MOCKED_UPLOADS["results"][0]["fastq"]["r2"][
                    "destination_path"
                ],
            ]
        )
    )
    echo(
        "\t".join(
            [
                MOCKED_UPLOADS["results"][1]["client_id"],
                MOCKED_UPLOADS["results"][1]["fastq"]["r1"]["upload"],
                MOCKED_UPLOADS["results"][1]["fastq"]["r1"][
                    "destination_path"
                ],
            ]
        )
    )
    assert output_line.getvalue() == res.output.encode()



"""Constants for upload command."""
from collections import namedtuple

from gencove.constants import Optionals

# pylint: disable=invalid-name

TMP_UPLOADS_WARNING = (
    "The Gencove upload area is a temporary storage. "
    "Log into the dashboard and assign uploaded files "
    "to a project in order to avoid automatic deletion."
)

_UploadStatuses = namedtuple("UploadStatuses", ["done", "started", "failed"])
UPLOAD_STATUSES = _UploadStatuses("succeeded", "started", "failed")

FASTQ_EXTENSIONS = (".fastq.gz", ".fastq.bgz", ".fq.gz", ".fq.bgz")

UPLOAD_PREFIX = "gncv://"

UploadOptions = namedtuple(
    "UploadOptions", Optionals._fields + ("project_id", "metadata")
)

ASSIGN_ERROR = (
    "Your files were successfully uploaded, "
    "but there was an error automatically running them "
    "and assigning them to project id {}."
    "You can try to assign without upload using following gncv path: {}"
)

ASSIGN_BATCH_SIZE = 200

FastQ = namedtuple("FastQ", ["client_id", "r_notation", "path"])

R_NOTATION_MAP = {"R1": "R1", "R2": "R2", "r1": "R1", "r2": "R2"}

_PathTemplateParts = namedtuple(
    "PathTemplateParts", ["client_id", "r_notation"]
)
PathTemplateParts = _PathTemplateParts("client_id", "r_notation")
PATH_TEMPLATE = "{{{}}}_{{{}}}.fastq.gz".format(
    PathTemplateParts.client_id, PathTemplateParts.r_notation
)



"""Test project's batches list command."""

import io
import sys
from uuid import uuid4

from click import echo
from click.testing import CliRunner

from gencove.client import APIClient  # noqa: I100
from gencove.command.projects.cli import list_project_batches


def test_list_project_batches__empty(mocker):
    """Test project has not batches."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_project_batches = mocker.patch.object(
        APIClient,
        "get_project_batches",
        return_value=dict(results=[], meta=dict(next=None)),
    )
    res = runner.invoke(
        list_project_batches,
        [str(uuid4()), "--email", "foo@bar.com", "--password", "123"],
    )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_get_project_batches.assert_called_once()
    assert res.output == ""


MOCKED_BATCHES = dict(
    meta=dict(count=2, next=None),
    results=[
        {
            "id": str(uuid4()),
            "name": "foo",
            "batch_type": "hd777k",
            "files": [
                {
                    "id": str(uuid4()),
                    "file_type": "report-zip",
                    "download_url": "https://foo.com/bar.zip",
                }
            ],
            "last_status": {
                "created": "2020-07-27T12:46:22.719862Z",
                "id": str(uuid4()),
                "status": "succeeded",
            },
        },
        {
            "id": str(uuid4()),
            "name": "bar",
            "batch_type": "hd777k",
            "files": [
                {
                    "id": str(uuid4()),
                    "file_type": "report-zip",
                    "download_url": "https://baz.com/bar.zip",
                }
            ],
            "last_status": {
                "created": "2020-07-28T12:46:22.719862Z",
                "id": str(uuid4()),
                "status": "running",
            },
        },
    ],
)


def test_list_project_batches__not_empty(mocker):
    """Test project batches being outputed to the shell."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_project_batches = mocker.patch.object(
        APIClient, "get_project_batches", return_value=MOCKED_BATCHES
    )

    res = runner.invoke(
        list_project_batches,
        [str(uuid4()), "--email", "foo@bar.com", "--password", "123"],
    )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_get_project_batches.assert_called_once()

    output_line = io.BytesIO()
    sys.stdout = output_line
    line_one = "\t".join(
        [
            MOCKED_BATCHES["results"][0]["id"],
            MOCKED_BATCHES["results"][0]["last_status"]["created"],
            MOCKED_BATCHES["results"][0]["last_status"]["status"],
            MOCKED_BATCHES["results"][0]["batch_type"],
            MOCKED_BATCHES["results"][0]["name"],
        ]
    )
    line_two = "\t".join(
        [
            MOCKED_BATCHES["results"][1]["id"],
            MOCKED_BATCHES["results"][1]["last_status"]["created"],
            MOCKED_BATCHES["results"][1]["last_status"]["status"],
            MOCKED_BATCHES["results"][1]["batch_type"],
            MOCKED_BATCHES["results"][1]["name"],
        ]
    )
    echo("\n".join([line_one, line_two]))
    assert output_line.getvalue() == res.output.encode()



"""Test project's batch types list command."""

import io
import sys
from uuid import uuid4

from click import echo
from click.testing import CliRunner

from gencove.client import APIClient  # noqa: I100
from gencove.command.projects.cli import list_project_batch_types


def test_list_project_batch_types__empty(mocker):
    """Test project has not batch types."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_project_batch_types = mocker.patch.object(
        APIClient,
        "get_project_batch_types",
        return_value=dict(results=[], meta=dict(next=None)),
    )
    res = runner.invoke(
        list_project_batch_types,
        [str(uuid4()), "--email", "foo@bar.com", "--password", "123"],
    )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_get_project_batch_types.assert_called_once()
    assert res.output == ""


MOCKED_BATCH_TYPES = dict(
    meta=dict(count=2, next=None),
    results=[
        {
            "key": "hd777k",
            "description": "foo",
        },
        {
            "key": "illuminahd",
            "description": "bar",
        },
    ],
)


def test_list_project_batch_types__not_empty(mocker):
    """Test project batch types being outputed to the shell."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_project_batch_types = mocker.patch.object(
        APIClient, "get_project_batch_types", return_value=MOCKED_BATCH_TYPES
    )

    res = runner.invoke(
        list_project_batch_types,
        [str(uuid4()), "--email", "foo@bar.com", "--password", "123"],
    )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_get_project_batch_types.assert_called_once()

    output_line = io.BytesIO()
    sys.stdout = output_line
    line_one = "\t".join(
        [
            MOCKED_BATCH_TYPES["results"][0]["key"],
            MOCKED_BATCH_TYPES["results"][0]["description"],
        ]
    )
    line_two = "\t".join(
        [
            MOCKED_BATCH_TYPES["results"][1]["key"],
            MOCKED_BATCH_TYPES["results"][1]["description"],
        ]
    )
    echo("\n".join([line_one, line_two]))
    assert output_line.getvalue() == res.output.encode()



"""Test project's restore samples command."""

from uuid import uuid4

from click.testing import CliRunner

from gencove.client import APIClient, APIClientError  # noqa: I100
from gencove.command.projects.cli import restore_project_samples


def test_restore_project_samples__bad_project_id(mocker):
    """Test restore project samples when non-uuid string is used as project
    id."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_restore_project_samples = mocker.patch.object(
        APIClient,
        "restore_project_samples",
    )
    res = runner.invoke(
        restore_project_samples,
        [
            "1111111",
            "--email",
            "foo@bar.com",
            "--password",
            "123",
            "--sample-ids",
            "11111111-1111-1111-1111-111111111111,22222222-2222-2222-2222-222222222222",  # noqa
        ],
    )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_restore_project_samples.assert_not_called()
    assert "Project ID is not valid" in res.output


def test_restore_project_samples__not_owned_project(mocker):
    """Test restore project samples failure when project is not owned."""
    mocked_response = {"detail": "Not found."}

    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_restore_project_samples = mocker.patch.object(
        APIClient, "restore_project_samples", return_value=mocked_response
    )

    mocked_restore_project_samples = mocker.patch.object(
        APIClient,
        "restore_project_samples",
        side_effect=APIClientError(message="", status_code=404),
    )

    res = runner.invoke(
        restore_project_samples,
        [
            str(uuid4()),
            "--email",
            "foo@bar.com",
            "--password",
            "123",
            "--sample-ids",
            "11111111-1111-1111-1111-111111111111,22222222-2222-2222-2222-222222222222",  # noqa
        ],
    )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_restore_project_samples.assert_called_once()
    assert "you do not have permission required" in res.output


def test_restore_project_samples__empty_sample_ids(mocker):
    """Test restore project samples failure when an empty list of sample ids
    is sent."""
    mocked_response = {"sample_ids": ["This list may not be empty."]}

    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_restore_project_samples = mocker.patch.object(
        APIClient, "restore_project_samples", return_value=mocked_response
    )

    mocked_restore_project_samples = mocker.patch.object(
        APIClient,
        "restore_project_samples",
        side_effect=APIClientError(message="", status_code=400),
    )

    res = runner.invoke(
        restore_project_samples,
        [
            str(uuid4()),
            "--email",
            "foo@bar.com",
            "--password",
            "123",
            "--sample-ids",
            "",
        ],
    )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_restore_project_samples.assert_not_called()
    assert "Missing sample IDs" in res.output


def test_restore_project_samples__invalid_sample_ids(mocker):
    """Test restore project samples failure when an empty list of sample ids
    is sent."""
    mocked_response = {"sample_ids": ["This list may not be empty."]}

    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_restore_project_samples = mocker.patch.object(
        APIClient, "restore_project_samples", return_value=mocked_response
    )

    mocked_restore_project_samples = mocker.patch.object(
        APIClient,
        "restore_project_samples",
        side_effect=APIClientError(message="", status_code=400),
    )

    res = runner.invoke(
        restore_project_samples,
        [
            str(uuid4()),
            "--email",
            "foo@bar.com",
            "--password",
            "123",
            "--sample-ids",
            "1111,222",
        ],
    )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_restore_project_samples.assert_not_called()
    assert "Not all sample IDs are valid" in res.output


def test_restore_project_samples__sample_not_in_project(mocker):
    """Test restore project samples with sample not in project."""
    mocked_response = {
        "sample_ids": [
            "All sample ids must be part of the current project and "
            "in archived status."
        ]
    }

    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_restore_project_samples = mocker.patch.object(
        APIClient, "restore_project_samples", return_value=mocked_response
    )
    mocked_restore_project_samples = mocker.patch.object(
        APIClient,
        "restore_project_samples",
        side_effect=APIClientError(message="", status_code=400),
    )

    res = runner.invoke(
        restore_project_samples,
        [
            str(uuid4()),
            "--email",
            "foo@bar.com",
            "--password",
            "123",
            "--sample-ids",
            "11111111-1111-1111-1111-111111111111",
        ],
    )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_restore_project_samples.assert_called_once()
    assert (
        "There was an error requesting project samples restore" in res.output
    )


def test_restore_project_samples__success(mocker):
    """Test restore project samples success."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_restore_project_samples = mocker.patch.object(
        APIClient, "restore_project_samples", return_value=None
    )

    res = runner.invoke(
        restore_project_samples,
        [
            str(uuid4()),
            "--email",
            "foo@bar.com",
            "--password",
            "123",
            "--sample-ids",
            "11111111-1111-1111-1111-111111111111,22222222-2222-2222-2222-222222222222",  # noqa
        ],
    )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_restore_project_samples.assert_called_once()
    assert "Request to restore samples accepted" in res.output



"""Utilities for processing and validating samples."""


def get_line(sample):
    """Build a list of relevant data to be printed.

    Args:
        sample (dict): an object from project samples

    Returns:
        list(str): list of relevant data to be printed
    """
    return "\t".join(
        [
            sample["last_status"]["created"],
            sample["id"],
            sample["client_id"],
            sample["last_status"]["status"],
            sample["archive_last_status"]["status"],
        ]
    )



"""Describe constants in samples subcommand."""
import re
from collections import namedtuple

from gencove.constants import Optionals, SAMPLE_ARCHIVE_STATUS, SAMPLE_STATUS

SamplesOptions = namedtuple(  # pylint: disable=invalid-name
    "SamplesOptions",
    Optionals._fields + ("status", "archive_status", "search"),
)

ALLOWED_STATUSES_RE = re.compile(
    "|".join(
        ["{}".format(status) for status in SAMPLE_STATUS._asdict().values()]
    ),
    re.IGNORECASE,
)

ALLOWED_ARCHIVE_STATUSES_RE = re.compile(
    "|".join(
        [
            "{}".format(status)
            for status in SAMPLE_ARCHIVE_STATUS._asdict().values()
        ]
    ),
    re.IGNORECASE,
)



"""Describe constants in samples subcommand."""
import re
from collections import namedtuple

from gencove.constants import Optionals, SAMPLE_ASSIGNMENT_STATUS

UploadsOptions = namedtuple(  # pylint: disable=invalid-name
    "UploadsOptions", Optionals._fields + ("status", "search")
)

ALLOWED_STATUSES_RE = re.compile(
    "|".join(
        [
            "{}".format(status)
            for status in SAMPLE_ASSIGNMENT_STATUS._asdict().values()
        ]
    ),
    re.IGNORECASE,
)



"""Project batches list shell command definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials, Optionals

from .main import ListBatches


@click.command("list-batches")
@click.argument("project_id")
@add_options(common_options)
def list_project_batches(project_id, host, email, password, api_key):
    """List batches that are available for a project."""
    ListBatches(
        project_id, Credentials(email, password, api_key), Optionals(host)
    ).run()



"""Test project's batches list command."""

from uuid import uuid4

from click.testing import CliRunner

from gencove.client import APIClient  # noqa: I100
from gencove.command.projects.cli import get_batch


def test_get_batch__empty(mocker):
    """Test project has not batches."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    batch_id = str(uuid4())
    mocked_get_batch = mocker.patch.object(
        APIClient,
        "get_batch",
        return_value=dict(
            id=batch_id,
            name="cli-test-1",
            batch_type="batch-type-1",
            sample_ids=["sample-id-1", "sample-id-2"],
            last_status=dict(
                id="last-status-id-1",
                status="running",
                created="2020-08-02T22:13:54.547167Z",
            ),
            files=[],
        ),
    )

    res = runner.invoke(
        get_batch, [batch_id, "--email", "foo@bar.com", "--password", "123"]
    )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_get_batch.assert_called_once()
    assert (
        res.output == "ERROR: There are no deliverables available for batch"
        " {}.\nAborted!\n".format(batch_id)
    )


def test_get_batch__not_empty(mocker):
    """Test project batches being outputed to the shell."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        mocked_login = mocker.patch.object(
            APIClient, "login", return_value=None
        )
        batch_id = str(uuid4())
        mocked_get_batch = mocker.patch.object(
            APIClient,
            "get_batch",
            return_value=dict(
                id=batch_id,
                name="cli-test-1",
                batch_type="batch-type-1",
                sample_ids=["sample-id-1", "sample-id-2"],
                last_status=dict(
                    id="last-status-id-1",
                    status="running",
                    created="2020-08-02T22:13:54.547167Z",
                ),
                files=[
                    dict(
                        id="file-id-1",
                        s3_path=(
                            "output/simple_vcf2finalreport/job-id-1/"
                            "simple_vcf2finalreport/report.zip"
                        ),
                        size=None,
                        download_url=(
                            "https://bucket.s3.amazonaws.com/output/"
                            "simple_vcf2finalreport/job-id-1/"
                            "simple_vcf2finalreport/report.zip"
                        ),
                        file_type="report-zip",
                    )
                ],
            ),
        )
        mocked_download_file = mocker.patch(
            "gencove.command.projects.get_batch.main.download.utils."
            "download_file"
        )
        res = runner.invoke(
            get_batch,
            [
                batch_id,
                "--email",
                "foo@bar.com",
                "--password",
                "123",
                "--output-filename",
                "test.zip",
            ],
        )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_get_batch.assert_called_once()
    mocked_download_file.assert_called_once_with(
        "test.zip",
        "https://bucket.s3.amazonaws.com/output/simple_vcf2finalreport/"
        "job-id-1/simple_vcf2finalreport/report.zip",
        no_progress=False,
    )


def test_get_batch__no_progress_not_empty(mocker):
    """Test project batches being outputed to the shell without progress."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        mocked_login = mocker.patch.object(
            APIClient, "login", return_value=None
        )
        batch_id = str(uuid4())
        mocked_get_batch = mocker.patch.object(
            APIClient,
            "get_batch",
            return_value=dict(
                id=batch_id,
                name="cli-test-1",
                batch_type="batch-type-1",
                sample_ids=["sample-id-1", "sample-id-2"],
                last_status=dict(
                    id="last-status-id-1",
                    status="running",
                    created="2020-08-02T22:13:54.547167Z",
                ),
                files=[
                    dict(
                        id="file-id-1",
                        s3_path=(
                            "output/simple_vcf2finalreport/job-id-1/"
                            "simple_vcf2finalreport/report.zip"
                        ),
                        size=None,
                        download_url=(
                            "https://bucket.s3.amazonaws.com/output/"
                            "simple_vcf2finalreport/job-id-1/"
                            "simple_vcf2finalreport/report.zip"
                        ),
                        file_type="report-zip",
                    )
                ],
            ),
        )
        mocked_download_file = mocker.patch(
            "gencove.command.projects.get_batch.main.download.utils."
            "download_file"
        )
        res = runner.invoke(
            get_batch,
            [
                batch_id,
                "--email",
                "foo@bar.com",
                "--password",
                "123",
                "--output-filename",
                "test.zip",
                "--no-progress",
            ],
        )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_get_batch.assert_called_once()
    mocked_download_file.assert_called_once_with(
        "test.zip",
        "https://bucket.s3.amazonaws.com/output/simple_vcf2finalreport/"
        "job-id-1/simple_vcf2finalreport/report.zip",
        no_progress=True,
    )



"""Gencove API client.

Exclude imports from linters due to install aliases breaking the rules.
"""

import datetime
import json
import time
from builtins import str as text  # noqa
from urllib.parse import parse_qs, urljoin, urlparse

from requests import (  # pylint: disable=W0622
    ConnectTimeout,
    ConnectionError,
    ReadTimeout,
    get,
    post,
)

from gencove import constants  # noqa: I100
from gencove.constants import (
    SAMPLES_SHEET_SORT_BY,
    SAMPLE_ARCHIVE_STATUS,
    SAMPLE_ASSIGNMENT_STATUS,
    SAMPLE_SORT_BY,
    SAMPLE_STATUS,
    SORT_ORDER,
)
from gencove.logger import echo_debug


class DateTimeEncoder(json.JSONEncoder):
    """JSON encoder that knows how to encode datetime objects."""

    # pylint: disable=method-hidden
    def default(self, o):
        """Override default method of JSONEncoder."""
        if isinstance(o, datetime.datetime):
            return o.isoformat()
        return json.JSONEncoder.default(self, o)


class APIError(Exception):
    """Base API Error."""

    def __init__(self, message, status_code=None):
        super().__init__(message)
        self.message = message
        self.status_code = status_code


class APIClientError(APIError):
    """Base HTTP error."""


class APIClientTooManyRequestsError(APIClientError):
    """Too many requests (429) HTTP error."""

    status_code = 429


class APIClientTimeout(APIClientError):
    """API timeout exception.

    From http://docs.python-requests.org/en/master/user/quickstart/#timeouts

        timeout is not a time limit on the entire response download;
        rather, an exception is raised if the server
        has not issued a response for timeout seconds
        (more precisely, if no bytes have been received on the underlying
        socket for timeout seconds).
        If no timeout is specified explicitly, requests do not time out.

    Catching this error will catch both
    :exc:`~requests.exceptions.ConnectTimeout` and
    :exc:`~requests.exceptions.ReadTimeout` errors.

    ConnectTimeout:
        The request timed out while trying to connect to the remote server.
        Requests that produced this error are safe to retry.

    ReadTimeout:
        The server did not send any data in the allotted amount of time.
    """


# pylint: disable=too-many-public-methods
class APIClient:
    """Gencove API client."""

    endpoints = constants.API_ENDPOINTS

    def __init__(self, host=None):
        """Initialize api client."""
        self._jwt_token = None
        self._jwt_refresh_token = None
        self._api_key = None
        self.host = host if host is not None else constants.HOST

    @staticmethod
    def _serialize_post_payload(payload):
        return json.dumps(payload, cls=DateTimeEncoder)

    # pylint: disable=bad-option-value,bad-continuation,too-many-arguments
    # pylint: disable=too-many-branches
    def _request(
        self,
        endpoint="",
        params=None,
        method="get",
        custom_headers=None,
        timeout=60,
        sensitive=False,
    ):
        url = urljoin(text(self.host), text(endpoint))
        headers = {"content-type": "application/json", "date": None}
        if custom_headers:
            headers.update(custom_headers)

        if not params:
            params = {}

        echo_debug(
            "Contacting url: {} with payload: {}".format(
                url, "[SENSITIVE CONTENT]" if sensitive else params
            )
        )
        start = time.time()

        try:
            if method == "get":
                response = get(
                    url=url, params=params, headers=headers, timeout=timeout
                )
            else:
                post_payload = APIClient._serialize_post_payload(params)
                response = post(
                    url=url,
                    data=post_payload,
                    headers=headers,
                    timeout=timeout,
                )

            if response.status_code == 429:
                raise APIClientTooManyRequestsError("Too Many Requests")
        except (ConnectTimeout, ConnectionError):
            # If request timed out,
            # let upper level handle it the way it sees fit.
            # one place might want to retry another might not.
            raise APIClientTimeout(  # pylint: disable=W0707
                "Could not connect to the api server"
            )
        except ReadTimeout:
            raise APIClientTimeout(  # pylint: disable=W0707
                "API server did not respond in timely manner"
            )

        echo_debug(
            "API response is {} status is {} in {}ms".format(
                "[SENSITIVE CONTENT]" if sensitive else response.content,
                response.status_code,
                (time.time() - start) * 1000,
            )
        )

        # pylint: disable=no-member
        if response.status_code >= 200 and response.status_code < 300:
            return response.json() if response.text else {}

        http_error_msg = ""
        if 400 <= response.status_code < 500:
            http_error_msg = "API Client Error: {}".format(response.reason)
            if response.text:
                response_json = response.json()
                if "detail" in response_json:
                    http_error_msg += ": {}".format(response_json["detail"])
                else:
                    try:
                        error_msg = "\n".join(
                            [
                                # create-batch can return error details that
                                # is a dict, not a list
                                "  {}: {}".format(
                                    key,
                                    value[0]
                                    if isinstance(value, list)
                                    else str(value),
                                )
                                for key, value in response_json.items()
                            ]
                        )
                    except AttributeError:
                        error_msg = "\n".join(response_json)
                    http_error_msg += ":\n{}".format(error_msg)

        elif 500 <= response.status_code < 600:
            http_error_msg = "Server Error: {}".format(response.reason)

        raise APIClientError(http_error_msg, response.status_code)

    def _set_jwt(self, access_token, refresh_token=None):
        self._jwt_token = access_token
        if refresh_token is not None:
            self._jwt_refresh_token = refresh_token

    def _refresh_authentication(self):
        echo_debug("Refreshing authentication")
        jwt = self.refresh_token(self._jwt_refresh_token)
        self._set_jwt(jwt["access"])

    def _get_authorization(self):
        if self._api_key:
            return {"Authorization": "Api-Key {}".format(self._api_key)}
        return {"Authorization": "Bearer {}".format(self._jwt_token)}

    def _post(
        self,
        endpoint,
        payload=None,
        timeout=120,
        authorized=False,
        sensitive=False,
        refreshed=False,
    ):
        headers = {} if not authorized else self._get_authorization()
        try:
            return self._request(
                endpoint,
                params=payload,
                method="post",
                timeout=timeout,
                custom_headers=headers,
                sensitive=sensitive,
            )
        except APIClientError as err:
            if not refreshed and err.status_code and err.status_code == 401:
                self._refresh_authentication()
                return self._post(
                    endpoint, payload, timeout, authorized, sensitive, True
                )

            raise err

    def _get(
        self,
        endpoint,
        query_params=None,
        timeout=120,
        authorized=False,
        sensitive=False,
        refreshed=False,
    ):
        headers = {} if not authorized else self._get_authorization()
        try:
            return self._request(
                endpoint,
                params=query_params,
                method="get",
                timeout=timeout,
                custom_headers=headers,
                sensitive=sensitive,
            )
        except APIClientError as err:
            if not refreshed and err.status_code and err.status_code == 401:
                self._refresh_authentication()
                return self._get(
                    endpoint,
                    query_params,
                    timeout,
                    authorized,
                    sensitive,
                    True,
                )

            raise err

    @staticmethod
    def _add_query_params(next_link, query_params=None, limit=200):
        if not query_params:
            query_params = {}
        query_params.update({"offset": 0, "limit": limit})
        if next_link:
            query_params["offset"] = parse_qs(urlparse(next_link).query)[
                "offset"
            ]
        return query_params

    def set_api_key(self, api_key):
        """Set api key on this instance."""
        self._api_key = api_key

    def refresh_token(self, refresh_token):
        """Refresh jwt token."""
        return self._post(
            self.endpoints.refresh_jwt,
            {"refresh": refresh_token},
            sensitive=True,
        )

    def validate_token(self, token):
        """Validate jwt token."""
        return self._post(
            self.endpoints.verify_jwt, {"token": token}, sensitive=True
        )

    def get_jwt(self, email, password):
        """Get jwt token."""
        return self._post(
            self.endpoints.get_jwt,
            dict(email=email, password=password),
            sensitive=True,
        )

    def login(self, email, password):
        """Log user in."""
        jwt = self.get_jwt(email, password)
        self._set_jwt(jwt["access"], jwt["refresh"])

    def get_upload_details(self, gncv_file_path):
        """Get file upload details.
        Args:
            gncv_file_path (str): gncv notated path

        Returns:
            dict: Upload details from the backend
        """
        return self._post(
            self.endpoints.upload_details,
            dict(destination_path=gncv_file_path),
            authorized=True,
        )

    def get_upload_credentials(self):
        """Get aws credentials for user."""
        return self._post(
            self.endpoints.get_upload_credentials,
            authorized=True,
            sensitive=True,
        )

    def get_project_samples(
        self,
        project_id,
        next_link=None,
        search="",
        sample_status=SAMPLE_STATUS.all,
        sample_archive_status=SAMPLE_ARCHIVE_STATUS.all,
        sort_by=SAMPLE_SORT_BY.modified,
        sort_order=SORT_ORDER.desc,
    ):
        """List single project's associated samples."""
        project_endpoint = self.endpoints.project_samples.format(
            id=project_id
        )
        params = self._add_query_params(
            next_link,
            {
                "search": search,
                "sort_by": sort_by,
                "sort_order": sort_order,
                "status": sample_status,
                "archive_status": sample_archive_status,
            },
        )
        return self._get(
            project_endpoint, query_params=params, authorized=True
        )

    def add_samples_to_project(self, samples, project_id, metadata=None):
        """Assign samples to a project.

        Args:
            samples (list of dicts): sample sheet results
            project_id (str): project to which to assign the samples
            metadata (str): JSON metadata to be applied to all samples
        """
        payload = {"uploads": samples, "metadata": metadata}

        return self._post(
            self.endpoints.project_samples.format(id=project_id),
            payload,
            authorized=True,
        )

    def get_sample_details(self, sample_id):
        """Fetch single sample details."""
        return self._get(
            self.endpoints.sample_details.format(id=sample_id),
            authorized=True,
        )

    def get_sample_qc_metrics(self, sample_id):
        """Fetch sample qc metrics.

        Args:
            sample_id(str of uuid): gencove sample id

        Returns:
            list of qc params objects.
        """
        return self._get(
            self.endpoints.sample_qc_metrics.format(id=sample_id),
            authorized=True,
        )

    def get_sample_sheet(
        self,
        gncv_path=None,
        assigned_status=SAMPLE_ASSIGNMENT_STATUS.all,
        next_link=None,
        sort_by=SAMPLES_SHEET_SORT_BY.created,
        sort_order=SORT_ORDER.desc,
    ):
        """Fetch user samples.

        Args:
            gncv_path (str): filter samples by gncv notated path.
            assigned_status (str, optional, default 'all'): filter samples by
                assignment status. One of SAMPLE_ASSIGNMENT_STATUS.
            next_link (str, optional): url from previous
                response['meta']['next'].
            sort_by(str): upload fastq field to sort by
            sort_order(str): asc or desc sorting
        """
        params = self._add_query_params(
            next_link,
            {
                "search": gncv_path,
                "status": assigned_status,
                "sort_by": sort_by,
                "sort_order": sort_order,
            },
        )
        return self._get(
            self.endpoints.sample_sheet, query_params=params, authorized=True
        )

    def list_projects(self, next_link=None):
        """Fetch projects.

        Args:
            next_link (str, optional): url from previous
                response['meta']['next'].

        Returns:
            api response (dict):
                {
                    "meta": {
                        "count": int,
                        "next": str,
                        "previous": optional[str],
                    },
                    "results": [...]
                }
        """
        params = self._add_query_params(next_link)
        return self._get(
            self.endpoints.projects, query_params=params, authorized=True
        )

    def get_pipeline_capabilities(self, pipeline_id):
        """Get pipeline capabilities details.

        Args:
            pipeline_id (str:uuid): pipeline id

        Returns:
            dict: pipeline details
        """
        resp = self._get(
            self.endpoints.pipeline_capabilities.replace("{id}", pipeline_id),
            authorized=True,
        )
        return resp

    def get_project_batch_types(self, project_id, next_link=None):
        """List single project's available batch types."""
        project_endpoint = self.endpoints.project_batch_types.format(
            id=project_id
        )
        params = self._add_query_params(next_link)
        return self._get(
            project_endpoint, query_params=params, authorized=True
        )

    def get_project_batches(self, project_id, next_link=None):
        """List single project's batches."""
        project_endpoint = self.endpoints.project_batches.format(
            id=project_id
        )
        params = self._add_query_params(next_link)
        return self._get(
            project_endpoint, query_params=params, authorized=True
        )

    def create_project_batch(
        self, project_id, batch_type, batch_name, sample_ids
    ):
        """Making a post request to create project batch."""
        project_endpoint = self.endpoints.project_batches.format(
            id=project_id
        )

        payload = {
            "name": batch_name,
            "batch_type": batch_type,
            "sample_ids": sample_ids,
        }

        return self._post(project_endpoint, payload, authorized=True)

    def get_batch(self, batch_id):
        """Get single batch."""
        batches_endpoint = self.endpoints.batches.format(id=batch_id)
        return self._get(batches_endpoint, authorized=True)

    def restore_project_samples(self, project_id, sample_ids):
        """Make a request to restore samples in given project."""
        restore_project_samples_endpoint = (
            self.endpoints.project_restore_samples.format(id=project_id)
        )

        payload = {"sample_ids": sample_ids}

        return self._post(
            restore_project_samples_endpoint, payload, authorized=True
        )

    def get_project(self, project_id):
        """Get single project."""
        project_endpoint = "{}{}".format(self.endpoints.projects, project_id)
        return self._get(project_endpoint, authorized=True)

    def create_merged_vcf(self, project_id):
        """Merge VCF files for a project."""
        merge_vcf_endpoint = self.endpoints.project_merge_vcfs.format(
            id=project_id
        )
        return self._post(merge_vcf_endpoint, authorized=True)

    def retrieve_merged_vcf(self, project_id):
        """Retrieve the status of the merge command for a project."""
        merge_vcf_endpoint = self.endpoints.project_merge_vcfs.format(
            id=project_id
        )
        return self._get(merge_vcf_endpoint, authorized=True)

    def get_metadata(self, sample_id):
        """Retrieve the metadata for a sample."""
        sample_metadata_endpoint = self.endpoints.sample_metadata.format(
            id=sample_id
        )
        return self._get(sample_metadata_endpoint, authorized=True)

    def set_metadata(self, sample_id, metadata):
        """Assign the metadata to a sample."""
        sample_metadata_endpoint = self.endpoints.sample_metadata.format(
            id=sample_id
        )
        payload = {
            "metadata": metadata,
        }
        return self._post(sample_metadata_endpoint, payload, authorized=True)



"""Project batch types list shell command definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials, Optionals

from .main import ListBatchTypes


@click.command("list-batch-types")
@click.argument("project_id")
@add_options(common_options)
def list_project_batch_types(project_id, host, email, password, api_key):
    """List batch types that are available for a project."""
    ListBatchTypes(
        project_id, Credentials(email, password, api_key), Optionals(host)
    ).run()



"""Describe constants in projects subcommand."""
from collections import namedtuple

Project = namedtuple(
    "Project",
    [
        "id",
        "name",
        "description",
        "created",
        "organization",
        "webhook_url",
        "sample_count",
        "pipeline_capabilities",
    ],
)
PipelineCapabilities = namedtuple(
    "PipelineCapabilities", ["id", "name", "private", "merge_vcfs_enabled"]
)



"""Test project list samples command."""
import io
import sys
from datetime import datetime, timedelta
from uuid import uuid4

from click.testing import CliRunner

from gencove.client import APIClient, APIClientError
from gencove.command.projects.cli import list_project_samples
from gencove.logger import echo_data


def test_list_empty(mocker):
    """Test project has no samples."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_project_samples = mocker.patch.object(
        APIClient,
        "get_project_samples",
        return_value=dict(results=[], meta=dict(next=None)),
    )
    res = runner.invoke(
        list_project_samples,
        [str(uuid4()), "--email", "foo@bar.com", "--password", "123"],
    )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_get_project_samples.assert_called_once()
    assert res.output == ""


def test_list_projects_bad_project_id(mocker):
    """Test project samples throw an error if project id is not uuid."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)

    project_id = "123456"

    res = runner.invoke(
        list_project_samples,
        [project_id, "--email", "foo@bar.com", "--password", "123"],
    )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    output_line = io.BytesIO()
    sys.stdout = output_line
    echo_data(
        "\n".join(
            [
                "ERROR: Project ID is not valid. Exiting.",
                "Aborted!",
            ]
        )
    )
    assert output_line.getvalue() == res.output.encode()


def test_list_projects_no_project(mocker):
    """Test project samples throw an error if no project available."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_project_samples = mocker.patch.object(
        APIClient,
        "get_project_samples",
        side_effect=APIClientError(
            message="API Client Error: Not Found: Not found.", status_code=404
        ),
        return_value={"detail": "Not found"},
    )

    project_id = str(uuid4())

    res = runner.invoke(
        list_project_samples,
        [project_id, "--email", "foo@bar.com", "--password", "123"],
    )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_get_project_samples.assert_called_once()

    output_line = io.BytesIO()
    sys.stdout = output_line
    echo_data(
        "\n".join(
            [
                "ERROR: Project {} does not exist or you do not have"
                " permission required to access it.".format(project_id),
                "ERROR: API Client Error: Not Found: Not found.",
                "Aborted!",
            ]
        )
    )
    assert output_line.getvalue() == res.output.encode()


def test_list_project_samples(mocker):
    """Test project samples being outputed to the shell."""
    mocked_samples = dict(
        meta=dict(count=1, next=None),
        results=[
            {
                "id": str(uuid4()),
                "client_id": "tester client id",
                "last_status": {
                    "created": (
                        datetime.utcnow() - timedelta(days=3)
                    ).isoformat(),
                    "status": "succeeded",
                },
                "archive_last_status": {
                    "status": "available",
                    "created": (
                        datetime.utcnow() - timedelta(days=1)
                    ).isoformat(),
                    "transition_cutoff": (
                        datetime.utcnow() + timedelta(days=6)
                    ).isoformat(),
                },
            }
        ],
    )
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_project_samples = mocker.patch.object(
        APIClient, "get_project_samples", return_value=mocked_samples
    )

    res = runner.invoke(
        list_project_samples,
        [str(uuid4()), "--email", "foo@bar.com", "--password", "123"],
    )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_get_project_samples.assert_called_once()

    output_line = io.BytesIO()
    sys.stdout = output_line
    echo_data(
        "\t".join(
            [
                mocked_samples["results"][0]["last_status"]["created"],
                mocked_samples["results"][0]["id"],
                mocked_samples["results"][0]["client_id"],
                mocked_samples["results"][0]["last_status"]["status"],
                mocked_samples["results"][0]["archive_last_status"]["status"],
            ]
        )
    )
    assert output_line.getvalue() == res.output.encode()



"""Create project's batch executor."""
import backoff

import requests

from gencove import client  # noqa: I100
from gencove.command.base import Command
from gencove.command.projects.list_batches.utils import get_line
from gencove.command.utils import is_valid_uuid
from gencove.exceptions import ValidationError


class CreateBatch(Command):
    """Create project's batch executor."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        project_id,
        batch_type,
        batch_name,
        sample_ids,
        credentials,
        options,
    ):
        super().__init__(credentials, options)
        self.project_id = project_id
        self.batch_type = batch_type
        self.batch_name = batch_name
        self.sample_ids = sample_ids

    def initialize(self):
        """Initialize list subcommand."""
        self.login()

    def validate(self):
        """Validate command input.

        Raises:
            ValidationError - if something is wrong with command parameters.
        """
        if not self.batch_type:
            raise ValidationError(
                "You must provide value for --batch-type. Exiting."
            )

        if not self.batch_name:
            raise ValidationError(
                "You must provide value for --batch-name. Exiting."
            )

        if is_valid_uuid(self.project_id) is False:
            raise ValidationError("Project ID is not valid. Exiting.")

        if self.sample_ids:
            if not all(is_valid_uuid(s_id) for s_id in self.sample_ids):
                raise ValidationError(
                    "Not all sample IDs are valid. Exiting."
                )

    @backoff.on_exception(
        backoff.expo,
        (requests.exceptions.ConnectionError, requests.exceptions.Timeout),
        max_tries=5,
        max_time=30,
    )
    def execute(self):
        """Make a request to create a batch for given project."""
        self.echo_debug(
            "Creating batch for project {} and batch key {}".format(
                self.project_id, self.batch_name
            )
        )

        try:
            created_batches_details = self.api_client.create_project_batch(
                project_id=self.project_id,
                batch_type=self.batch_type,
                batch_name=self.batch_name,
                sample_ids=self.sample_ids,
            )
            self.echo_debug(created_batches_details)
            for batch in created_batches_details["results"]:
                self.echo_data(get_line(batch))
        except client.APIClientError as err:
            self.echo_debug(err)
            if err.status_code == 400:
                self.echo_warning(
                    "There was an error creating project batches."
                )
                self.echo_info("The following error was returned:")
                self.echo_info(err.message)
            elif err.status_code == 404:
                self.echo_warning(
                    "Project {} does not exist or you do not have "
                    "permission required to access it.".format(
                        self.project_id
                    )
                )
            else:
                raise



"""Describe all constants in Gencove CLI."""
from collections import namedtuple

HOST = "https://api.gencove.com"

_APIEndpoint = namedtuple(
    "ApiEndpoint",
    [
        "get_jwt",
        "refresh_jwt",
        "verify_jwt",
        "upload_details",
        "get_upload_credentials",
        "project_samples",
        "sample_details",
        "sample_qc_metrics",
        "sample_sheet",
        "projects",
        "pipeline_capabilities",
        "project_batch_types",
        "project_batches",
        "batches",
        "project_merge_vcfs",
        "sample_metadata",
        "project_restore_samples",
    ],
)
API_ENDPOINTS = _APIEndpoint(
    "/api/v2/jwt-create/",
    "/api/v2/jwt-refresh/",
    "/api/v2/jwt-verify/",
    "/api/v2/uploads-post-data/",
    "/api/v2/upload-credentials/",
    "/api/v2/project-samples/{id}",
    "/api/v2/samples/{id}",
    "/api/v2/sample-quality-controls/{id}",
    "/api/v2/sample-sheet/",
    "/api/v2/projects/",
    "/api/v2/pipeline-capabilities/{id}",
    "/api/v2/project-batch-types/{id}",
    "/api/v2/project-batches/{id}",
    "/api/v2/batches/{id}",
    "/api/v2/project-merge-vcfs/{id}",
    "/api/v2/sample-metadata/{id}",
    "/api/v2/project-restore-samples/{id}",
)

_SampleAssignmentStatus = namedtuple(
    "SampleAssignmentStatus", ["all", "unassigned", "assigned"]
)
SAMPLE_ASSIGNMENT_STATUS = _SampleAssignmentStatus(
    "all", "unassigned", "assigned"
)

_SampleSheetSortBy = namedtuple("SampleSheetSortBy", ["created", "modified"])
SAMPLES_SHEET_SORT_BY = _SampleSheetSortBy("created", "modified")

_SampleStatus = namedtuple(
    "SampleStatus", ["completed", "succeeded", "failed", "running", "all"]
)
SAMPLE_STATUS = _SampleStatus(
    "completed", "succeeded", "failed", "running", "all"
)

_SampleArchiveStatus = namedtuple(
    "SampleArchiveStatus",
    [
        "available",
        "archived",
        "restore_requested",
        "all",
    ],
)

SAMPLE_ARCHIVE_STATUS = _SampleArchiveStatus(
    "available",
    "archived",
    "restore_requested",
    "all",
)

_SampleSortBy = namedtuple(
    "SampleSortFields", ["created", "modified", "status", "client_id", "id"]
)
SAMPLE_SORT_BY = _SampleSortBy(
    "created", "modified", "status", "client_id", "id"
)

_SortOrder = namedtuple("SortOrder", ["asc", "desc"])
SORT_ORDER = _SortOrder("asc", "desc")

Credentials = namedtuple("Credentials", ["email", "password", "api_key"])
Optionals = namedtuple("Optionals", ["host"])

_DownloadTemplateParts = namedtuple(
    "DownloadTemplateParts",
    [
        "client_id",
        "gencove_id",
        "file_type",
        "file_extension",
        "default_filename",
    ],
)
# pylint: disable=C0103
DownloadTemplateParts = _DownloadTemplateParts(
    "client_id",
    "gencove_id",
    "file_type",
    "file_extension",
    "default_filename",
)
DOWNLOAD_TEMPLATE = "{{{}}}/{{{}}}/{{{}}}".format(
    DownloadTemplateParts.client_id,
    DownloadTemplateParts.gencove_id,
    DownloadTemplateParts.default_filename,
)

MAX_RETRY_TIME_SECONDS = 300  # 5 minutes
FASTQ_MAP_EXTENSION = ".fastq-map.csv"



"""Project status merged VCF shell command definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials, Optionals

from .main import StatusMergedVCF


@click.command("status-merged-vcf")
@click.argument("project_id")
@add_options(common_options)
def status_merged_vcf(
    project_id,
    host,
    email,
    password,
    api_key,
):
    """Get status of merge VCF files job in a project."""
    StatusMergedVCF(
        project_id,
        Credentials(email, password, api_key),
        Optionals(host),
    ).run()



"""List uploads subcommand."""
import backoff

import requests

# pylint: disable=wrong-import-order
from gencove.client import APIClientError  # noqa: I100
from gencove.command.base import Command
from gencove.command.utils import validate_input
from gencove.constants import SAMPLE_ASSIGNMENT_STATUS

from .constants import ALLOWED_STATUSES_RE
from .utils import get_line


class ListSampleSheet(Command):
    """List sample sheet command executor."""

    def __init__(self, credentials, options):
        super().__init__(credentials, options)
        self.status = options.status
        self.gncv_path = options.search

    def initialize(self):
        """Initialize list subcommand."""
        self.login()

    def validate(self):
        """Validate command input."""
        validate_input(
            "upload status",
            self.status,
            ALLOWED_STATUSES_RE,
            SAMPLE_ASSIGNMENT_STATUS,
        )

    def execute(self):
        self.echo_debug(
            "Retrieving sample sheet: "
            "status={} search_term={}".format(self.status, self.gncv_path)
        )
        try:
            for uploads in self.get_paginated_sample_sheet():
                if not uploads:
                    self.echo_debug("No matching uploads found.")
                    return

                for upload in uploads:
                    self.echo_data(get_line(upload))
        except APIClientError as err:
            if err.status_code == 404:
                self.echo_error(
                    "Uploads do not exist or you do not have "
                    "permission required to access them."
                )
            raise

    def get_paginated_sample_sheet(self):
        """Paginate over all sample sheet pages.

        Yields:
            paginated lists of uploads
        """
        more = True
        next_link = None
        while more:
            self.echo_debug("Get sample sheet page")
            resp = self.get_sample_sheet(next_link)
            yield resp["results"]
            next_link = resp["meta"]["next"]
            more = next_link is not None

    @backoff.on_exception(
        backoff.expo,
        (requests.exceptions.ConnectionError, requests.exceptions.Timeout),
        max_tries=5,
        max_time=30,
    )
    def get_sample_sheet(self, next_link=None):
        """Get sample sheet page."""
        return self.api_client.get_sample_sheet(
            gncv_path=self.gncv_path,
            assigned_status=self.status,
            next_link=next_link,
        )



"""Utilities for processing and validating uploads from sample sheet."""


def get_line(upload):
    """Build a list of relevant data to be printed.

    Args:
        upload (dict): object from sample sheet

    Returns:
        list(str): list of relevant data to be printed
    """
    parts = [
        upload["client_id"],
        upload["fastq"]["r1"]["upload"],
        upload["fastq"]["r1"]["destination_path"],
    ]
    if "r2" in upload["fastq"]:
        parts.append(upload["fastq"]["r2"]["upload"])
        parts.append(upload["fastq"]["r2"]["destination_path"])

    return "\t".join(parts)



"""Python library which enables you to use Gencoves' research backend."""
import click

from gencove import version
from gencove.command.download import download
from gencove.command.projects import projects
from gencove.command.samples import samples
from gencove.command.upload import upload
from gencove.command.uploads import uploads


@click.group()
@click.version_option(version=version.version())
def cli():
    """Gencove's command line interface."""


cli.add_command(download)
cli.add_command(upload)
cli.add_command(uploads)
cli.add_command(projects)
cli.add_command(samples)


if __name__ == "__main__":
    cli()



"""Download command constants."""
import re
from collections import namedtuple

from gencove.constants import DownloadTemplateParts, Optionals


_SampleStatuses = namedtuple("SampleStatuses", ["succeeded", "failed"])
_SampleArchiveStatuses = namedtuple(
    "SampleStatuses", ["available", "restored"]
)
SAMPLE_STATUSES = _SampleStatuses("succeeded", "failed")
SAMPLE_ARCHIVE_STATUSES = _SampleArchiveStatuses("available", "restored")

ALLOWED_STATUSES_RE = re.compile(
    "{}|{}".format(SAMPLE_STATUSES.succeeded, SAMPLE_STATUSES.failed),
    re.IGNORECASE,
)
ALLOWED_ARCHIVE_STATUSES_RE = re.compile(
    "{}|{}".format(
        SAMPLE_ARCHIVE_STATUSES.available, SAMPLE_ARCHIVE_STATUSES.restored
    ),
    re.IGNORECASE,
)
KILOBYTE = 1024
MEGABYTE = 1024 * KILOBYTE
NUM_MB_IN_CHUNK = 3
CHUNK_SIZE = NUM_MB_IN_CHUNK * MEGABYTE

DownloadFilters = namedtuple(
    "Filters", ["project_id", "sample_ids", "file_types"]
)
DownloadOptions = namedtuple(  # pylint: disable=invalid-name
    "DownloadOptions",
    Optionals._fields + ("skip_existing", "download_template"),
)
DEFAULT_FILENAME_TOKEN = "{{{}}}".format(
    DownloadTemplateParts.default_filename
)
FilePrefix = namedtuple(
    "FilePrefix",
    ["dirs", "filename", "file_extension", "use_default_filename"],
)
FILENAME_RE = re.compile("filename=(.+)")

# Leaving the line below since it is a great explaination of FILE_TYPES_MAPPER
# usage:
# FILE_TYPES_MAPPER = {"fastq-r1": "fastq-r1_R1", "fastq-r2": "fastq-r2_R2"}
FILE_TYPES_MAPPER = {}

QC_FILE_TYPE = "qc"
METADATA_FILE_TYPE = "metadata"



"""Samples list shell command definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials, SAMPLE_ASSIGNMENT_STATUS

from .constants import UploadsOptions
from .main import ListSampleSheet


@click.command("list")
@click.option("--search", help="Filter uploads by gncv path")
@click.option(
    "--status",
    help="Filter uploads by status of assignment",
    type=click.Choice(SAMPLE_ASSIGNMENT_STATUS._asdict().values()),
    default=SAMPLE_ASSIGNMENT_STATUS.all,
)
@add_options(common_options)
def list_uploads(  # pylint: disable=E0012,C0330,R0913
    search, status, host, email, password, api_key
):
    """List user uploads."""
    ListSampleSheet(
        Credentials(email, password, api_key),
        UploadsOptions(host, status, search),
    ).run()



"""Test project's batches create command."""

import io
import sys
from uuid import uuid4

from click import echo
from click.testing import CliRunner

from gencove.client import APIClient, APIClientError  # noqa: I100
from gencove.command.projects.cli import create_project_batch


def test_create_project_batches__missing_batch_type(mocker):
    """Test batch creation failure when batch type is not sent."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_create_project_batch = mocker.patch.object(
        APIClient,
        "create_project_batch",
    )
    res = runner.invoke(
        create_project_batch,
        [
            str(uuid4()),
            "--email",
            "foo@bar.com",
            "--password",
            "123",
            "--batch-name",
            "foo bar",
        ],
    )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_create_project_batch.assert_not_called()
    assert "You must provide value for --batch-type" in res.output


def test_create_project_batches__missing_batch_name(mocker):
    """Test batch creation failure when batch name is not sent."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_create_project_batch = mocker.patch.object(
        APIClient,
        "create_project_batch",
    )
    res = runner.invoke(
        create_project_batch,
        [
            str(uuid4()),
            "--email",
            "foo@bar.com",
            "--password",
            "123",
            "--batch-type",
            "hd777k",
        ],
    )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_create_project_batch.assert_not_called()
    assert "You must provide value for --batch-name" in res.output


def test_create_project_batches__bad_project_id(mocker):
    """Test batch creation failure when non-uuid string is used as project
    id.
    """
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_create_project_batch = mocker.patch.object(
        APIClient,
        "create_project_batch",
    )
    res = runner.invoke(
        create_project_batch,
        [
            "1111111",
            "--email",
            "foo@bar.com",
            "--password",
            "123",
            "--batch-type",
            "hd777k",
            "--batch-name",
            "foo bar",
        ],
    )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_create_project_batch.assert_not_called()
    assert "Project ID is not valid" in res.output


def test_create_project_batches__not_owned_project(mocker):
    """Test batch creation failure when project is not owned."""
    mocked_response = {"detail": "Not found."}

    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_create_project_batch = mocker.patch.object(
        APIClient, "create_project_batch", return_value=mocked_response
    )

    mocked_create_project_batch = mocker.patch.object(
        APIClient,
        "create_project_batch",
        side_effect=APIClientError(message="", status_code=404),
    )

    res = runner.invoke(
        create_project_batch,
        [
            str(uuid4()),
            "--email",
            "foo@bar.com",
            "--password",
            "123",
            "--batch-type",
            "hd777k",
            "--batch-name",
            "foo bar",
        ],
    )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_create_project_batch.assert_called_once()
    assert "you do not have permission required" in res.output


def test_create_project_batches__duplicate_client_ids(mocker):
    """Test batch creation failure when there are samples that share same
    client id.
    """
    mocked_response = {
        "sample_ids": ["All samples must have unique client_id attribute."],
        "duplicate_client_ids": {
            "cow1": [
                "85b3d2c4-7215-4fb0-9cc6-5731bae71ea2",
                "01dc9ad3-f05d-40bb-967f-715e03cb2108",
            ]
        },
    }

    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_create_project_batch = mocker.patch.object(
        APIClient, "create_project_batch", return_value=mocked_response
    )

    mocked_create_project_batch = mocker.patch.object(
        APIClient,
        "create_project_batch",
        side_effect=APIClientError(message="", status_code=400),
    )

    res = runner.invoke(
        create_project_batch,
        [
            str(uuid4()),
            "--email",
            "foo@bar.com",
            "--password",
            "123",
            "--batch-type",
            "hd777k",
            "--batch-name",
            "foo bar",
        ],
    )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_create_project_batch.assert_called_once()
    assert "There was an error creating project batches" in res.output


def test_create_project_batches__success__with_sample_ids(mocker):
    """Test batch creation success when when sample ids are explicitly sent."""
    mocked_response = {
        "meta": {"count": 1},
        "results": [
            {
                "id": str(uuid4()),
                "name": "foo bar",
                "batch_type": "hd777k",
                "last_status": {
                    "created": "2020-07-28T12:46:22.719862Z",
                    "id": str(uuid4()),
                    "status": "running",
                },
            }
        ],
    }

    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_create_project_batch = mocker.patch.object(
        APIClient, "create_project_batch", return_value=mocked_response
    )

    res = runner.invoke(
        create_project_batch,
        [
            str(uuid4()),
            "--email",
            "foo@bar.com",
            "--password",
            "123",
            "--batch-type",
            "hd777k",
            "--batch-name",
            "foo bar",
            "--sample-ids",
            "11111111-1111-1111-1111-111111111111,22222222-2222-2222-2222-222222222222",  # noqa
        ],
    )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_create_project_batch.assert_called_once()

    output_line = io.BytesIO()
    sys.stdout = output_line
    echo(
        "\t".join(
            [
                mocked_response["results"][0]["id"],
                mocked_response["results"][0]["last_status"]["created"],
                mocked_response["results"][0]["last_status"]["status"],
                mocked_response["results"][0]["batch_type"],
                mocked_response["results"][0]["name"],
            ]
        )
    )
    assert output_line.getvalue() == res.output.encode()


def test_create_project_batches__success__without_sample_ids(mocker):
    """Test batch creation success when when sample ids are not explicitly
    sent.
    """
    mocked_response = {
        "meta": {"count": 1},
        "results": [
            {
                "id": str(uuid4()),
                "name": "foo bar",
                "batch_type": "hd777k",
                "last_status": {
                    "created": "2020-07-28T12:46:22.719862Z",
                    "id": str(uuid4()),
                    "status": "running",
                },
            }
        ],
    }

    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_create_project_batch = mocker.patch.object(
        APIClient, "create_project_batch", return_value=mocked_response
    )

    res = runner.invoke(
        create_project_batch,
        [
            str(uuid4()),
            "--email",
            "foo@bar.com",
            "--password",
            "123",
            "--batch-type",
            "hd777k",
            "--batch-name",
            "foo bar",
        ],
    )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_create_project_batch.assert_called_once()

    output_line = io.BytesIO()
    sys.stdout = output_line
    echo(
        "\t".join(
            [
                mocked_response["results"][0]["id"],
                mocked_response["results"][0]["last_status"]["created"],
                mocked_response["results"][0]["last_status"]["status"],
                mocked_response["results"][0]["batch_type"],
                mocked_response["results"][0]["name"],
            ]
        )
    )
    assert output_line.getvalue() == res.output.encode()



"""Common utils used in multiple commands."""
import uuid

from gencove.exceptions import ValidationError


def sanitize_string(output):
    """Removes unwanted characters from output string."""
    return output.replace("\t", " ")


def validate_input(key, provided_value, allowed_values_re, allowed_values):
    """Validates provided value is in allowed values.

    Args:
        key(str): key for which the provided value is being validated
        provided_value (str): one of sort order/sort by field/sample status
        allowed_values_re(re.compiled value): values to compare against,
            compiled via re module
        allowed_values(namedtuple): of valid values for the key

    Returns:
        None: if everything is valid

    Raises:
         ValidationError: with validation error message
    """
    if not allowed_values_re.match(provided_value):
        raise ValidationError(
            "Unknown {} value: {}. Allowed values are: {}".format(
                key,
                provided_value,
                ", ".join(allowed_values._asdict().values()),
            )
        )


def is_valid_uuid(candidate):
    """Test if provided string is a valid uuid version 4 string.

    candidate (str): uuid to check

    Returns:
        bool: True if is a valid uuid v4, False if not
    """
    try:
        uuid.UUID(candidate, version=4)
        return True
    except ValueError:
        return False
